///////////////////////////////////////////////////////////////////////////////
// File		:	GLModel.cpp
// Creator	:	Ashley Bennett
// Date		:	Saturday 11 September 2005
// Purpose	:	A class to load and display MRF 3D models.
///////////////////////////////////////////////////////////////////////////////

#include "GLModel.h"
#include "MyXML.h"				//To Load in files
//#include "TextureFunctions.h"	//For loading textures
#include <Math3d.h>				//For m3dNormalizeVectorf()

bool LoadTextureBMP(const char *szFileName, unsigned int unTexID);
bool LoadTextureTGA(const char *szFileName, unsigned int unTexID);

//Default Constructor
GLModel::GLModel()
{
	m_vdata			= 0;
	m_ndata			= 0;
	m_fdata			= 0;
	m_uvdata		= 0;
	m_unNumVerts	= 0;
	m_unNumFaces	= 0;
	m_unNumTextures	= 0;
	m_unTextureID = 0;
	ZeroMemory(m_szTexture, _MAX_PATH);
}

//Default Destructor
GLModel::~GLModel()
{
	if (m_vdata)
		delete m_vdata;
	if (m_ndata)
		delete m_ndata;
	if (m_fdata)
		delete m_fdata;
	if (m_uvdata)
		delete m_uvdata;

	resetTextureData();
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	resetVertData
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Deletes any old data in the vertex data array, and resizes the
// array to the length in m_unNumVerts.
///////////////////////////////////////////////////////////////////////////////
void GLModel::resetVertData(void)
{
	//Delete old data
	if (m_vdata)
		delete m_vdata;

	//Create Rows (Vertex * 3 (X, Y and Z))
	m_vdata = new GLfloat[m_unNumVerts * 3];

	//Clear to 0's
	memset(m_vdata, 0, m_unNumVerts * 3 * sizeof(GLfloat));
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	resetNormalData
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Deletes any old data in the normal data array, and resizes the
// array to the length in m_unNumVerts.
///////////////////////////////////////////////////////////////////////////////
void GLModel::resetNormalData(void)
{
	//Delete old data
	if (m_ndata)
		delete m_ndata;

	//Create Rows (Normal * 3 (X, Y and Z))
	m_ndata = new GLfloat[m_unNumVerts * 3];

	//Clear to 0's
	memset(m_ndata, 0, m_unNumVerts * 3 * sizeof(GLfloat));
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	resetFaceData
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Deletes any old data in the face data array, and resizes the
// array to the length in m_unNumFaces.
///////////////////////////////////////////////////////////////////////////////
void GLModel::resetFaceData(void)
{
	//Delete old data
	if (m_fdata)
		delete m_fdata;

	//Create Rows (Face * 3 (A, B and C))
	m_fdata = new GLuint[m_unNumFaces * 3];

	//Clear to 0's
	memset(m_fdata, 0, m_unNumFaces * 3 * sizeof(GLint));
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	resetUVData
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Deletes any old data in the UV data array, and resizes the
// array to the length in m_unNumVerts.
///////////////////////////////////////////////////////////////////////////////
void GLModel::resetUVData(void)
{
	//Delete old data
	if (m_uvdata)
		delete m_uvdata;

	//Create Rows (Face * 2 (U and V))
	m_uvdata = new GLfloat[m_unNumVerts * 2];

	//Clear to 0's
	memset(m_uvdata, 0, m_unNumVerts * 2 * sizeof(GLuint));
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	resetTextureData
// Creator	:	Ashley Bennett
// Date		:	Monday 19 September 2005
// Purpose	:	Deletes clears the path to the model's texture and the number
// of textures.
///////////////////////////////////////////////////////////////////////////////
void GLModel::resetTextureData(void)
{
	//If a texture was loaded, free it
	if (m_unTextureID)	
		glDeleteTextures(1, &m_unTextureID);

	//Clear the texture path
	ZeroMemory(m_szTexture, _MAX_PATH);

	m_unNumTextures = 0;
	m_unTextureID = 0;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	loadTextures
// Creator	:	Ashley Bennett
// Date		:	Monday 19 September 2005
// Purpose	:	Loads in the textures for this model and sets up necessary
// variables, such as the Texture ID.
///////////////////////////////////////////////////////////////////////////////
void GLModel::loadTextures(void)
{
	//Get a handle to a texture
	glGenTextures(1, &m_unTextureID);

	//Check to see what type of texture we have a path to
	char* szExtension = &m_szTexture[strlen(m_szTexture) - 3];

	//Capture the result of the loading function, and cleanup if there is a failure
	bool bResult = false;

	if (!strcmp(szExtension, "bmp"))
	{
		bResult = LoadTextureBMP(m_szTexture, m_unTextureID);
	}
	else if (!strcmp(szExtension, "tga"))
	{
		bResult = LoadTextureTGA(m_szTexture, m_unTextureID);
	}

	if (!bResult)
		resetTextureData();
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	scale
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Scales the model by multiplying the verticies by the amount
// passed into this function.
///////////////////////////////////////////////////////////////////////////////
void GLModel::scale(float fScale)
{
	for (UINT i = 0; i < m_unNumVerts * 3; i++)
		m_vdata[i] *= fScale;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	flipNormals
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Turns the normals to point the other way to make objects that
// are "inside-out" correct again.
///////////////////////////////////////////////////////////////////////////////
void GLModel::flipNormals(void)
{
	for (UINT i = 0; i < m_unNumVerts * 3; i++)
		m_ndata[i] *= -1;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	normalizeNormals
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Make all normals unit vectors to speed up processing.
///////////////////////////////////////////////////////////////////////////////
void GLModel::normalizeNormals(void)
{
	for (UINT i = 0; i < m_unNumVerts * 3; i += 3)
		m3dNormalizeVectorf(&m_ndata[i]);
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	loadXML
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Sets up a model by loading in data from an XML file.
///////////////////////////////////////////////////////////////////////////////
void GLModel::loadXML(char *szPath)
{
	// Cleanup current data
	resetTextureData();

	// Create an instance of the XML parser	
	ISAXXMLReader    *pRdr = NULL;
	MyXML			  cHandler;
	wchar_t           wcBuffer[ 1024 ];

	// COM setup (try SAXXMLReader if SAXXMLReader40 fails)
	CoInitialize(NULL); 
	CoCreateInstance( __uuidof(SAXXMLReader40), NULL, CLSCTX_ALL, __uuidof(ISAXXMLReader), (void**)&pRdr );

	pRdr->putContentHandler( &cHandler );

	// Build wchar string of file
	mbstowcs( wcBuffer, szPath, 1023 );

	// Pass this model as the model to load into
	cHandler.setGLModel(this);

	// Start parsing
	pRdr->parseURL( wcBuffer );
	pRdr->Release();

	CoUninitialize();
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	loadBIN
// Creator	:	Ashley Bennett
// Date		:	Sunday 12 September 2005
// Purpose	:	Sets up a model by loading in data from an BIN file.
///////////////////////////////////////////////////////////////////////////////
void GLModel::loadBIN(char *szPath)
{
	// Cleanup current data
	resetTextureData();

	// create an output stream to save the file
	ifstream fin;
	fin.open(szPath, ios_base::binary);

	/*
	// length of the string we are about to read in
	unsigned int unLen = 0;

	// read in the intro
	//fin.read( reinterpret_cast<char*>(&unLen), sizeof(unLen) );
	//fin.read( meshname, unLen );

	//fin.read( reinterpret_cast<char*>(&unLen), sizeof(unLen) );
	//fin.read( dagPath, unLen );

	fin.read( reinterpret_cast<char*>(&m_unNumVerts), sizeof(m_unNumVerts) );
	fin.read( reinterpret_cast<char*>(&m_unNumFaces), sizeof(m_unNumFaces) );
	fin.read( reinterpret_cast<char*>(&m_unNumTextures), sizeof(m_unNumTextures) );

	// setup the arrays to read into
	resetVertData();
	resetNormalData();
	resetFaceData();
	resetUVData();

	// read in all the textures
	for (unsigned int i = 0; i < m_unNumTextures; i++)
	{
		//fin.read( reinterpret_cast<char*>(&unLen), sizeof(unLen) );
		//fin.read( texturename, unLen );
	}

	//Variables to read peices of data into
	float fData = 0;
	double dData = 0;
	UINT unData = 0;

	// read in the vertices, normals and UVs
	for (unsigned int i = 0, j = 0; i < m_unNumVerts * 3; i += 3, j += 2)
	{
		fin.read( reinterpret_cast<char*>(&dData), sizeof(double) );
		m_vdata[i] = dData;
		fin.read( reinterpret_cast<char*>(&dData), sizeof(double) );
		m_vdata[i+1] = dData;
		fin.read( reinterpret_cast<char*>(&dData), sizeof(double) );
		m_vdata[i+2] = dData;

		fin.read( reinterpret_cast<char*>(&fData), sizeof(float) );
		m_ndata[i] = fData;
		fin.read( reinterpret_cast<char*>(&fData), sizeof(float) );
		m_ndata[i+1] = fData;
		fin.read( reinterpret_cast<char*>(&fData), sizeof(float) );
		m_ndata[i+2] = fData;

		fin.read( reinterpret_cast<char*>(&unData), sizeof(int) );
		m_uvdata[j] = unData;
		fin.read( reinterpret_cast<char*>(&unData), sizeof(int) );
		m_uvdata[j+1] = unData;

		//fin.read( reinterpret_cast<char*>(&m_vdata[i]), sizeof(GLfloat) );
		//fin.read( reinterpret_cast<char*>(&m_vdata[i+1]), sizeof(GLfloat) );
		//fin.read( reinterpret_cast<char*>(&m_vdata[i+2]), sizeof(GLfloat) );
		//fin.read( reinterpret_cast<char*>(&m_ndata[i]), sizeof(GLfloat) );
		//fin.read( reinterpret_cast<char*>(&m_ndata[i+1]), sizeof(GLfloat) );
		//fin.read( reinterpret_cast<char*>(&m_ndata[i+2]), sizeof(GLfloat) );
		//fin.read( reinterpret_cast<char*>(&m_uvdata[j]), sizeof(GLint) );
		//fin.read( reinterpret_cast<char*>(&m_uvdata[j+1]), sizeof(GLint) );
	}

	// read in the faces
	for (unsigned int i = 0; i < m_unNumFaces * 3; i += 3)
	{
		fin.read( reinterpret_cast<char*>(&m_fdata[i]), sizeof(GLint) );
		fin.read( reinterpret_cast<char*>(&m_fdata[i+1]), sizeof(GLint) );
		fin.read( reinterpret_cast<char*>(&m_fdata[i+2]), sizeof(GLint) );
	}
	*/

	// read in the sizes of the arrays
	fin.read( reinterpret_cast<char*>(&m_unNumVerts), sizeof(m_unNumVerts) );
	fin.read( reinterpret_cast<char*>(&m_unNumFaces), sizeof(m_unNumFaces) );
	fin.read( reinterpret_cast<char*>(&m_unNumTextures), sizeof(m_unNumTextures) );

	// setup the arrays to read into
	resetVertData();
	resetNormalData();
	resetFaceData();
	resetUVData();

	// read in the arrays
	fin.read( reinterpret_cast<char*>(m_vdata),  m_unNumVerts * 3 * sizeof(GLfloat) );
	fin.read( reinterpret_cast<char*>(m_ndata),  m_unNumVerts * 3 * sizeof(GLfloat) );
	fin.read( reinterpret_cast<char*>(m_uvdata), m_unNumVerts * 2 * sizeof(GLfloat) );
	fin.read( reinterpret_cast<char*>(m_fdata),  m_unNumFaces * 3 * sizeof(GLuint) );

	// close file
	fin.close();
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	draw()
// Creator	:	Ashley Bennett
// Date		:	Saturday 11 September 2005
// Purpose	:	Draw a 3D model loaded from an MRF.
///////////////////////////////////////////////////////////////////////////////
void GLModel::draw(void)
{
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, m_vdata);

	glEnableClientState(GL_NORMAL_ARRAY);
	glNormalPointer(GL_FLOAT, 0, m_ndata);

	//If this model has its own textures, use the UV's
	if (m_unNumTextures)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 0, m_uvdata);
	}

	glDrawElements(GL_TRIANGLES, m_unNumFaces * 3, GL_UNSIGNED_INT, m_fdata);

	if (m_unNumTextures)
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}
