///////////////////////////////////////////////////////////////////////////////
// File		:	Game.h
// Creator	:	Ashley Bennett
// Date		:	Saturday 11 September 2005
// Purpose	:	Imports an MRF and displays the model.
///////////////////////////////////////////////////////////////////////////////

#pragma once

//---INCLUDES---

#include <windows.h>
#include <cmath>
#include <stdio.h>		//For sprintf() used in FPS counter
#include <direct.h>		//For _getcwd() for accessing assets

#include <gl/gl.h>		//OpenGL 1.1
#include <gl/glu.h>		//GLU (OpenGL Helpers)
#include <gl/glext.h>	//OpenGL Extensions
#include <gl/extgl.h>	//OpenGL Extension Loading Library Header

#include <GLFrame.h>
#include <StopWatch.h>
#include <glFrustum.h>
#include "TextureFunctions.h"

#include "GLModel.h"


//---DEFINES---

#define CAM_MOVE		1.0f
#define CAM_ROTATE		1.0f

#if !defined(M_PI)
#define M_PI 3.14159265F
#endif

enum MOUSEEVENTS { M_NONE = 0, M_TRANSLATE, M_ROTATE, M_ZOOM };

//Keeps an angle between -360.0f and 360.0f
#define ClampAngle(x) x += (x > 360.0f) ? -360.0f : (x < -360.0f) ? 360.0f : 0

//Could have used GLVertex in Structures.h
struct Vector3f
{
	GLfloat x, y, z;
};


//---GLOBALS---

GLsizei		g_siWinWidth, g_siWinHeight;	//Width and height of the window
GLFrame		g_frCamera;						//A viewpoint to draw from
GLFrustum	g_Frustum;						//A culling frustum
GLModel		g_Model;						//A Model to load and draw
Vector3f	g_vRotate;						//Amount to Rotate the world
//Vector3f	g_vTranslate;					//Amount to Translate the world

POINT		g_ptMouse, g_ptOldMouse;		// Mouse position
UINT		g_unMouseState;					// How to process Mouse Move Events

HWND		hWnd;							//Handle to the window


///////////////////////////////////////////////////////////////////////////////
// Function	:	ResetWorld()
// Creator	:	Ashley Bennett
// Date		:	Monday 19 September 2005
// Purpose	:	Resets the translate and rotate values to apply to the world
// when rendering. You'll be moved back from the origin, staring down the Z.
///////////////////////////////////////////////////////////////////////////////
void ResetWorld(void)
{
	//g_vTranslate.x = g_vTranslate.y = 0; g_vTranslate.z = -20.0f;
	g_frCamera.SetOrigin(0.0f, 0.0f, 0.0f);
	g_frCamera.MoveForward(-20.0f);
	g_vRotate.x = g_vRotate.y = g_vRotate.z = 0;
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	GetInput()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	Gets keyboard input and affects objects on screen.
///////////////////////////////////////////////////////////////////////////////
void GetInput(void)
{
	//float fDeltaTime = g_Timer.GetElapsedSeconds();
	//float fDeltaTime = g_Timer.GetTime();
	float fDeltaTime = 1.0f;

	// Move Left and Right
	if ( GetAsyncKeyState('A') )
		g_frCamera.MoveRight(CAM_MOVE * fDeltaTime);
	else if ( GetAsyncKeyState('D') )
		g_frCamera.MoveRight(-CAM_MOVE * fDeltaTime);

	// Move Up and Down
	if ( GetAsyncKeyState('W') )
		g_frCamera.MoveUp(CAM_MOVE * fDeltaTime);
	else if ( GetAsyncKeyState('S') )
		g_frCamera.MoveUp(-CAM_MOVE * fDeltaTime);

	// Rotate Left and Right
	if ( GetAsyncKeyState(VK_LEFT) )
		g_frCamera.RotateLocalY( m3dDegToRad(CAM_ROTATE * fDeltaTime) );
	else if ( GetAsyncKeyState(VK_RIGHT) )
		g_frCamera.RotateLocalY( m3dDegToRad(-CAM_ROTATE * fDeltaTime) );

	// Move Forward and Back
	if ( GetAsyncKeyState(VK_UP) )
		g_frCamera.MoveForward(CAM_MOVE * fDeltaTime);
	else if ( GetAsyncKeyState(VK_DOWN) )
		g_frCamera.MoveForward(-CAM_MOVE * fDeltaTime);

	//Move the frustum with the camera
	g_Frustum.Transform(g_frCamera);
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	ShowFPS()
// Creator	:	Ashley Bennett
// Date		:	Sunday 18 September 2005
// Purpose	:	Shows the FPS in the window title bar.
///////////////////////////////////////////////////////////////////////////////
void ShowFPS(void)
{
	static CStopWatch fpsTimer;
	static int nFrames = 0;

	//Start the FPS timer once
	if (nFrames == 0)
		fpsTimer.Reset();

	//Increment Frame counter
	nFrames++;

	//Update the FPS every 100 frames
	if (nFrames == 101)
	{
		float fps = 100.0f / fpsTimer.GetElapsedSeconds();

		//Make a message of the FPS
		char szBuffer[64];
		sprintf(szBuffer, "OpenGL Model Viewer (FPS: %.1f)", fps);

		//Apply to the Window's Title Bar
		SetWindowText(hWnd, szBuffer);

		//Reset counter for next time
		nFrames = 1;
		fpsTimer.Reset();
	}
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	SetupGL()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	One Time Setup of OpenGL Rendering.
///////////////////////////////////////////////////////////////////////////////
void SetupGL(void)
{
	//---BASICS---

	//Set the color to clear as dark blue
	glClearColor(0.2f, 0.2f, 0.4f, 0.0f);

	//Turn on depth testing
	glEnable(GL_DEPTH_TEST);

	//Enable Backface Culling
	glEnable(GL_CULL_FACE);

	//Draw in wireframe mode. Default is solid (GL_FILL).
	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	//---LIGHTING---

	//Turn on lighting
	glEnable(GL_LIGHTING);

	//Save some light details
	GLfloat noLight[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
	GLfloat medLight[4] = { 0.3f, 0.3f, 0.3f, 1.0f };
	GLfloat someLight[4] = { 0.85f, 0.85f, 0.85f, 0.0f };
	GLfloat lightPos[4] = { 0.0f, 0.0f, 0.0f, 1.0f };

	//Turn on light 0
	glEnable(GL_LIGHT0);

	//Set some of Light 0's properties
	glLightfv(GL_LIGHT0, GL_AMBIENT, noLight);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, someLight);
	glLightfv(GL_LIGHT0, GL_SPECULAR, noLight);
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

	//Wherever glColor() calls are placed, interpret them as meterials
	glEnable(GL_COLOR_MATERIAL);

	//Apply those calls to diffuse as materials on front facing faces
	glColorMaterial(GL_FRONT, GL_DIFFUSE);


	//---OPTIONS---

	//Set our default texturing mode
	g_unTextureMode = TEX_BMP;

	//Load Textures
	LoadTextures();

	//Turn on texturing
	glEnable(GL_TEXTURE_2D);


	//---RESOURCES---

	//Load in a model
	//g_Model.loadXML("Assets/megaman.xml");
	g_Model.loadXML("Assets/TexturedCube.xml");
	//g_Model.loadBIN("Assets/cube.bin");

	//Scale the model
	//g_Model.scale(0.01f);
	g_Model.scale(8.0f);

	//Normalize the model's Normals
	g_Model.normalizeNormals();

	//Flip the model's Normals
	//g_Model.flipNormals();

	//If the model has textures, lets show them
	if (g_Model.getNumTextures())
		g_unTextureMode = TEX_MODEL;


	//---WORLD POSITION---

	ResetWorld();
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	ShutdownGL()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	Cleanup OpenGL Rendering.
///////////////////////////////////////////////////////////////////////////////
void ShutdownGL(void)
{
	//Cleanup textures
	glDeleteTextures(NUM_TEXTURES, g_unTextures);
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	Render()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	Renders the scene.
///////////////////////////////////////////////////////////////////////////////
void Render(void)
{
	//Show the FPS counter
	//ShowFPS();

	//Take user input to effect the objects
	//GetInput();

	//Clear color and depth buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Save world matrix
	glPushMatrix();
	{
		//Apply camera Matrix
		g_frCamera.ApplyCameraTransform();

		//Set drawing color to White
		glColor3f(1.0f, 1.0f, 1.0f);

		//Rotate the model
		glRotatef(g_vRotate.x, 1.0f, 0.0f, 0.0f);
		glRotatef(g_vRotate.y, 0.0f, 1.0f, 0.0f);
		glRotatef(g_vRotate.z, 0.0f, 0.0f, 1.0f);

		//Move the model
		//glTranslatef(g_vTranslate.x, g_vTranslate.y, g_vTranslate.z);

		if (g_unTextureMode == TEX_MODEL)
		{
			//---MODEL'S TEXTURE---

			//Make sure model has its own texture
			if (g_Model.getNumTextures())
			{
				//Be sure texturing is on
				glEnable(GL_TEXTURE_2D);

				//Use the model's texture
				glBindTexture(GL_TEXTURE_2D, g_Model.getTextureID());
			}

			g_Model.draw();
		}
		else if (g_unTextureMode == TEX_NONE)
		{
			//---NO TEXTURES---

			g_Model.draw();
		}
		else if (g_unTextureMode == TEX_CUBEMAPBMP || g_unTextureMode == TEX_CUBEMAPDDS)
		{
			//---CUBEMAP MODE---

			//Set the Texture coordinate generation mode for all axis
			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
			glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
			glEnable(GL_TEXTURE_GEN_S);
			glEnable(GL_TEXTURE_GEN_T);
			glEnable(GL_TEXTURE_GEN_R);

			//Use our Cube map textures
			glDisable(GL_TEXTURE_2D);
			glEnable(GL_TEXTURE_CUBE_MAP);
			glBindTexture(GL_TEXTURE_CUBE_MAP, g_unTextures[0]);

			g_Model.draw();

			//Stop using our Cube map textures
			glDisable(GL_TEXTURE_CUBE_MAP);
			glEnable(GL_TEXTURE_2D);

			//Stop generating coordinates
			glDisable(GL_TEXTURE_GEN_S);
			glDisable(GL_TEXTURE_GEN_T);
			glDisable(GL_TEXTURE_GEN_R);

		}
		else
		{
			//---SPHERE MAP---

			//Use our sphere map texture
			glBindTexture(GL_TEXTURE_2D, g_unTextures[0]);

			//Set the Texture coordinate generation mode for all axis
			glEnable(GL_TEXTURE_GEN_S);
			glEnable(GL_TEXTURE_GEN_T);
			glEnable(GL_TEXTURE_GEN_R);
			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
			glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);

			g_Model.draw();

			//Stop generating coordinates
			glDisable(GL_TEXTURE_GEN_S);
			glDisable(GL_TEXTURE_GEN_T);
			glDisable(GL_TEXTURE_GEN_R);
		}
	}
	//Restore world identity matrix
	glPopMatrix();

	//Reset the timer
	//g_Timer.Reset();
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	ChangeSize()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	Runs when the window changes size.
///////////////////////////////////////////////////////////////////////////////
void ChangeSize(int w, int h)
{
	//Prevent Divide by 0 error
	if (h == 0)
		h = 1;

	//Resize the viewport
	glViewport(0, 0, w, h);

	//We need to fix the projection, so get in the correct mode
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	//Resize the projection frustum
	g_Frustum.Set(35.0f, (float)w / (float)h, 1.0f, 100.0f);

	//Get ready to draw again
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	GameStartup()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	One time setup for the game.
///////////////////////////////////////////////////////////////////////////////
void GameStartup(void)
{
	//Save the window width and height
	g_siWinWidth	= 800;
	g_siWinHeight	= 600;

	//Reset/Start the timer
	//g_Timer.Reset();
}


///////////////////////////////////////////////////////////////////////////////
// Function	:	WorldUpdate()
// Creator	:	Ashley Bennett
// Date		:	Thursday 1 September 2005
// Purpose	:	Sets the translate and rotate values to apply to the world
// when rendering. Use this with mouse movements to rotate models for example.
///////////////////////////////////////////////////////////////////////////////
void WorldUpdate(void)
{
	int nDeltaX = g_ptOldMouse.x - g_ptMouse.x;
	int nDeltaY = g_ptMouse.y - g_ptOldMouse.y;

	switch(g_unMouseState)
	{
	case M_TRANSLATE:
		{
			float fX = nDeltaX / 100.0f;
			float fY = nDeltaY / 100.0f;
			g_frCamera.MoveRight(-fX);
			g_frCamera.MoveUp(fY);
		}
		break;
	case M_ROTATE:
		{
			g_vRotate.x += (nDeltaY * 180.0f) / 500.0f;
			g_vRotate.y -= (nDeltaX * 180.0f) / 500.0f;
			ClampAngle(g_vRotate.x);
			ClampAngle(g_vRotate.y);
		}
		break;
	case M_ZOOM:
		{
			float fZoom = -(nDeltaX + nDeltaY) / 100.0f;
			g_frCamera.MoveForward(fZoom);
		}
		break;
	}
}
