///////////////////////////////////////////////////////////////////////////////
// File		:	TextureFunctions.h
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Functions for loading in image files for texturing with.
///////////////////////////////////////////////////////////////////////////////

#pragma once

//---INCLUDES---

#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glprocs.h>
#include "nv_dds.h"		//For loading DDS files
#include <string.h>		//For strcat() in LoadTextures()


//---DEFINES---

#define NUM_TEXTURES	1

enum TEXTUREMODES { TEX_NONE = 0, TEX_BMP, TEX_COMPRESSEDBMP, TEX_TGA, TEX_COMPRESSEDTGA, TEX_DDS,
					TEX_CUBEMAPBMP, TEX_CUBEMAPDDS, TEX_MODEL };


//---GLOBALS---

GLuint		g_unTextures[NUM_TEXTURES];		//Texture object handles
GLuint		g_unTextureMode;				//Flag for how we should draw textures (BMP, Compressed BMP, DDS etc.)

char		g_szAppDir[_MAX_PATH];			//Stores directory of the application


//---PROTOTYPES---

GLbyte *gltLoadTGA(const char *szFileName, GLint *iWidth, GLint *iHeight, GLint *iComponents, GLenum *eFormat);


///////////////////////////////////////////////////////////////////////////////
// Function	:	gltReadBMPBits()
// Creator	:	Richard Wright
// Date		:	Thursday 8 September 2005
// Purpose	:
// This function opens the bitmap file given (szFileName), verifies that it is
// a 24bit .BMP file and loads the bitmap bits needed so that it can be used
// as a texture. The width and height of the bitmap are returned in nWidth and
// nHeight. The memory block allocated and returned must be deleted with delete [];
// The returned array is an 888 BGR texture (use GL_BGR_EXT or BGRtoRGB for
// glTextImage2D
///////////////////////////////////////////////////////////////////////////////
BYTE* gltReadBMPBits(const char *szFileName, int *nWidth, int *nHeight)
{
	HANDLE hFileHandle;
	BITMAPINFO *pBitmapInfo = NULL;
	unsigned long lInfoSize = 0;
	unsigned long lBitSize = 0;
	BYTE *pBits = NULL;					// Bitmaps bits

	// Open the Bitmap file
	hFileHandle = CreateFile(szFileName,GENERIC_READ,FILE_SHARE_READ,
		NULL,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,NULL);

	// Check for open failure (most likely file does not exist).
	if(hFileHandle == INVALID_HANDLE_VALUE)
		return NULL;

	// File is Open. Read in bitmap header information
	BITMAPFILEHEADER	bitmapHeader;
	DWORD dwBytes;
	ReadFile(hFileHandle,&bitmapHeader,sizeof(BITMAPFILEHEADER),	
		&dwBytes,NULL);

	// Check for a couple of simple errors	
	if(dwBytes != sizeof(BITMAPFILEHEADER))
		return FALSE;

	// Check format of bitmap file
	if(bitmapHeader.bfType != 'MB')
		return FALSE;

	// Read in bitmap information structure
	lInfoSize = bitmapHeader.bfOffBits - sizeof(BITMAPFILEHEADER);
	pBitmapInfo = (BITMAPINFO *) new BYTE[lInfoSize];
	ReadFile(hFileHandle,pBitmapInfo,lInfoSize,&dwBytes,NULL);

	if(dwBytes != lInfoSize)
	{
		delete [] pBitmapInfo;
		CloseHandle(hFileHandle);
		return FALSE;
	}

	// Save the size and dimensions of the bitmap
	*nWidth = pBitmapInfo->bmiHeader.biWidth;
	*nHeight = pBitmapInfo->bmiHeader.biHeight;
	lBitSize = pBitmapInfo->bmiHeader.biSizeImage;

	// If the size isn't specified, calculate it anyway	
	if(pBitmapInfo->bmiHeader.biBitCount != 24)
	{
		delete [] pBitmapInfo;
		return FALSE;
	}

	if(lBitSize == 0)
		lBitSize = (*nWidth *
		pBitmapInfo->bmiHeader.biBitCount + 7) / 8 *
		abs(*nHeight);

	// Allocate space for the actual bitmap
	delete [] pBitmapInfo;
	pBits = new BYTE[lBitSize];

	// Read in the bitmap bits, check for corruption
	if(!ReadFile(hFileHandle,pBits,lBitSize,&dwBytes,NULL) ||
		dwBytes != (sizeof(BYTE)*lBitSize))
		pBits = NULL;

	// Close the bitmap file now that we have all the data we need
	CloseHandle(hFileHandle);

	return pBits;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadTextureBMP()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in a BMP texture from the passed path.
///////////////////////////////////////////////////////////////////////////////
bool LoadTextureBMP(const char *szFileName, unsigned int unTexID)
{
	BYTE	*pBytes;
	int		nTexW, nTexH;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_2D, unTexID);

	//Read in the data
	pBytes = gltReadBMPBits(szFileName, &nTexW, &nTexH);

	//Check file loaded
	if (pBytes == NULL)
	{
		MessageBox(NULL, "Texture File Not Found (BMP)", "Error", MB_OK);
		return false;
	}

	//Setup some texture loading/drawing parameters
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR /*GL_LINEAR*/ /*GL_NEAREST*/);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, /*GL_LINEAR_MIPMAP_LINEAR*/ GL_LINEAR /*GL_NEAREST*/);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE /*GL_DECAL*/);

	//OpenGL will copy the data to the video card
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, nTexW, nTexH, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, pBytes);

	//Cleanup
	delete [] pBytes;

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadTextureBMPCompressed()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in a BMP texture from the passed path and compress it.
///////////////////////////////////////////////////////////////////////////////
bool LoadTextureBMPCompressed(const char *szFileName, unsigned int unTexID)
{
	BYTE	*pBytes;
	int		nTexW, nTexH;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_2D, unTexID);

	//Read in the data
	pBytes = gltReadBMPBits(szFileName, &nTexW, &nTexH);

	//Check file loaded
	if (pBytes == NULL)
	{
		MessageBox(NULL, "Texture File Not Found (BMP)", "Error", MB_OK);
		return false;
	}

	//Setup some texture loading/drawing parameters
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR /*GL_LINEAR*/ /*GL_NEAREST*/);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, /*GL_LINEAR_MIPMAP_LINEAR*/ GL_LINEAR /*GL_NEAREST*/);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE /*GL_DECAL*/);

	//OpenGL will copy the data to the video card
	glTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGB, nTexW, nTexH, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, pBytes);

	//Cleanup
	delete [] pBytes;

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadTextureTGA()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in a TGA texture from the passed path.
///////////////////////////////////////////////////////////////////////////////
bool LoadTextureTGA(const char *szFileName, unsigned int unTexID)
{
	GLint	iComponents;
	GLenum	eFormat;
	GLbyte* pglBits;
	int		nTexW, nTexH;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_2D, unTexID);

	//Read in the data
	pglBits = gltLoadTGA(szFileName, &nTexW, &nTexH, &iComponents, &eFormat);

	//Check file loaded
	if (pglBits == NULL)
	{
		MessageBox(NULL, "Texture File Not Found (TGA)", "Error", MB_OK);
		return false;
	}

	//Setup some texture loading/drawing parameters
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	//OpenGL will copy the data to the video card
	glTexImage2D(GL_TEXTURE_2D, 0, iComponents, nTexW, nTexH, 0, eFormat, GL_UNSIGNED_BYTE, pglBits);

	//Cleanup
	free(pglBits);

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadTextureTGACompressed()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in a TGA texture from the passed path and compress it.
///////////////////////////////////////////////////////////////////////////////
bool LoadTextureTGACompressed(const char *szFileName, unsigned int unTexID)
{
	GLint	iComponents;
	GLenum	eFormat;
	GLbyte* pglBits;
	int		nTexW, nTexH;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_2D, unTexID);

	//Read in the data
	pglBits = gltLoadTGA(szFileName, &nTexW, &nTexH, &iComponents, &eFormat);

	//Check file loaded
	if (pglBits == NULL)
	{
		MessageBox(NULL, "Texture File Not Found (TGA)", "Error", MB_OK);
		return false;
	}

	//Change iComponents to be compressed
	switch(iComponents)
	{
	case GL_RGB8:     // Most likely case
		iComponents = GL_COMPRESSED_RGB;
		break;
	case GL_RGBA8:
		iComponents = GL_COMPRESSED_RGBA;
		break;
	case GL_LUMINANCE8:
		iComponents = GL_COMPRESSED_LUMINANCE;
		break;
	};

	//Setup some texture loading/drawing parameters
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	//OpenGL will copy the data to the video card
	glTexImage2D(GL_TEXTURE_2D, 0, iComponents, nTexW, nTexH, 0, eFormat, GL_UNSIGNED_BYTE, pglBits);

	//Cleanup
	free(pglBits);

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadTextureDDS()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in a DDS texture from the passed path.
///////////////////////////////////////////////////////////////////////////////
bool LoadTextureDDS(const char *szFileName, unsigned int unTexID)
{
	CDDSImage image;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_2D, unTexID);

	//Read in the data
	image.load(szFileName);

	//Setup some texture loading/drawing parameters
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	//OpenGL will copy the data to the video card
	glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, image.format, 
		image[0].width, image[0].height, 0, image[0].size, image[0]);

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadCubeMapBMP()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in the BMP textures to make a cube map. Example usage:
//				char cubeMapBMPs[6][32] = {
//					{"Assets/Terrain_posx.bmp"}, {"Assets/Terrain_negx.bmp"},
//					{"Assets/Terrain_posy.bmp"}, {"Assets/Terrain_negy.bmp"},
//					{"Assets/Terrain_posz.bmp"}, {"Assets/Terrain_negz.bmp"} };
//				LoadCubeMapBMP(cubeMapBMPs, g_Textures[x]);
///////////////////////////////////////////////////////////////////////////////
bool LoadCubeMapBMP(const char szFileName[][_MAX_PATH], unsigned int unTexID)
{
	BYTE	*pBytes;
	int		nTexW, nTexH;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_CUBE_MAP, unTexID);

	//For each face
	for (UINT i = 0; i < 6; i++)
	{
		//Read in the data
		pBytes = gltReadBMPBits(szFileName[i], &nTexW, &nTexH);

		//Check file loaded
		if (pBytes == NULL)
		{
			MessageBox(NULL, "Cube Texture File Not Found (BMP)", "Error", MB_OK);
			return false;
		}

		//Setup some texture loading/drawing parameters
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		//OpenGL will copy the data to the video card
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGB, nTexW, nTexH, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, pBytes);

		//Cleanup
		delete [] pBytes;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadCubeMapBMP()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in the DDS textures to make a cube map. Example usage:
//				char cubeMapDDSs[6][32] = {
//					{"Assets/cube_face_posx.dds"}, {"Assets/cube_face_negx.dds"},
//					{"Assets/cube_face_posy.dds"}, {"Assets/cube_face_negy.dds"},
//					{"Assets/cube_face_posz.dds"}, {"Assets/cube_face_negz.dds"} };
//				LoadCubeMapDDS(cubeMapDDSs, g_Textures[x]);
///////////////////////////////////////////////////////////////////////////////
bool LoadCubeMapDDS(const char szFileName[][_MAX_PATH], unsigned int unTexID)
{
	CDDSImage image;

	//The ID for this texture dictates where to put it in our list
	glBindTexture(GL_TEXTURE_CUBE_MAP, unTexID);

	//For each face
	for (UINT i = 0; i < 6; i++)
	{
		//Read in the data
		image.load(szFileName[i]);

		//Setup some texture loading/drawing parameters
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		//OpenGL will copy the data to the video card
		glCompressedTexImage2DARB(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, image.format, 
			image[0].width, image[0].height, 0, image[0].size, image[0]);
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function	:	LoadTextures()
// Creator	:	Ashley Bennett
// Date		:	Thursday 15 September 2005
// Purpose	:	Loads in all textures based on the current texture mode.
///////////////////////////////////////////////////////////////////////////////
bool LoadTextures(void)
{
	//Cleanup textures
	glDeleteTextures(NUM_TEXTURES, g_unTextures);

	//Fill our array with however many texture object handles we need
	glGenTextures(NUM_TEXTURES, g_unTextures);

	//Create a temporary path variable to load from
	char szPath[_MAX_PATH] = { 0 };
	strcat(szPath, g_szAppDir);

	if (g_unTextureMode == TEX_BMP)
	{
		strcat(szPath, "/Assets/spheremap.bmp");
		LoadTextureBMP(szPath, g_unTextures[0]);
	}
	else if (g_unTextureMode == TEX_COMPRESSEDBMP)
	{
		strcat(szPath, "/Assets/spheremap.bmp");
		LoadTextureBMPCompressed(szPath, g_unTextures[0]);
	}
	else if (g_unTextureMode == TEX_TGA)
	{
		strcat(szPath, "/Assets/spheremap.tga");
		LoadTextureTGA(szPath, g_unTextures[0]);
	}
	else if (g_unTextureMode == TEX_COMPRESSEDTGA)
	{
		strcat(szPath, "/Assets/spheremap.tga");
		LoadTextureTGACompressed(szPath, g_unTextures[0]);
	}
	else if (g_unTextureMode == TEX_DDS)
	{
		strcat(szPath, "/Assets/spheremap.dds");
		LoadTextureDDS(szPath, g_unTextures[0]);
	}
	else if (g_unTextureMode == TEX_CUBEMAPBMP)
	{
		char cubeMapBMPs[6][_MAX_PATH] = { 0 };
		
		//Put the base path in each string
		strcat(cubeMapBMPs[0], szPath);
		strcat(cubeMapBMPs[1], szPath);
		strcat(cubeMapBMPs[2], szPath);
		strcat(cubeMapBMPs[3], szPath);
		strcat(cubeMapBMPs[4], szPath);
		strcat(cubeMapBMPs[5], szPath);

		//Put the extra peice of the path in for each image
		strcat(cubeMapBMPs[0], "/Assets/Terrain_posx.bmp");
		strcat(cubeMapBMPs[1], "/Assets/Terrain_negx.bmp");
		strcat(cubeMapBMPs[2], "/Assets/Terrain_posy.bmp");
		strcat(cubeMapBMPs[3], "/Assets/Terrain_negy.bmp");
		strcat(cubeMapBMPs[4], "/Assets/Terrain_posz.bmp");
		strcat(cubeMapBMPs[5], "/Assets/Terrain_negz.bmp");

		LoadCubeMapBMP(cubeMapBMPs, g_unTextures[0]);
	}
	else if (g_unTextureMode == TEX_CUBEMAPDDS)
	{
		char cubeMapDDSs[6][_MAX_PATH] = { 0 };
		
		//Put the base path in each string
		strcat(cubeMapDDSs[0], szPath);
		strcat(cubeMapDDSs[1], szPath);
		strcat(cubeMapDDSs[2], szPath);
		strcat(cubeMapDDSs[3], szPath);
		strcat(cubeMapDDSs[4], szPath);
		strcat(cubeMapDDSs[5], szPath);

		//Put the extra peice of the path in for each image
		strcat(cubeMapDDSs[0], "/Assets/cube_face_posx.dds");
		strcat(cubeMapDDSs[1], "/Assets/cube_face_negx.dds");
		strcat(cubeMapDDSs[2], "/Assets/cube_face_posy.dds");
		strcat(cubeMapDDSs[3], "/Assets/cube_face_negy.dds");
		strcat(cubeMapDDSs[4], "/Assets/cube_face_posz.dds");
		strcat(cubeMapDDSs[5], "/Assets/cube_face_negz.dds");

		LoadCubeMapDDS(cubeMapDDSs, g_unTextures[0]);
	}

	return true;
}
