///////////////////////////////////////////////////////////////////////////////
//
// Description:
// 
// Loads DDS images (DXTC1, DXTC3, DXTC5, RGB (888, 888X), and RGBA (8888) are
// supported) for use in OpenGL. Image is flipped when its loaded as DX images
// are stored with different coordinate system. If file has mipmaps and/or 
// cubemaps then these are loaded as well. 
//
// Each face in the image is stored in a list and can be access using the array
// operator (i.e. image[0]). If the image is not a cubemap then just the first
// item in the list is valid, if it is a cubemap then items 0 through 5 will be
// valid. The cubemap member variable is set to true when a cubemap is loaded.
//
// The mipmaps for each face are also stored in a list and can be accessed like 
// so: image[0].mipmaps[0] (which accesses the first mipmap of the first 
// image). To get the number of mipmaps call the size function of the mipmap 
// list like so: numMipmaps = image[0].mipmaps.size();
//
///////////////////////////////////////////////////////////////////////////////
//
// Sample usage
//
///////////////////////////////////////////////////////////////////////////////
//
// Loading a compressed texture:
//
// CDDSImage image;
// GLuint texobj;
//
// image.load("compressed.dds");
// 
// glGenTextures(1, &texobj);
// glEnable(GL_TEXTURE_2D);
// glBindTexture(GL_TEXTURE_2D, texobj);
//
// glCompressedTexImage2DARB(GL_TEXTURE_2D, 0, image.format, 
//     image[0].width, image[0].height, 0, image[0].size, image[0]);
//
// for (int i = 0; i < image[0].mipmaps.size(); i++)
// {
//     glCompressedTexImage2DARB(GL_TEXTURE_2D, i+1, image.format, 
//         image[0].mipmaps[i].width, image[0].mipmaps[i].height, 0, 
//         image[0].mipmaps[i].size, image[0].mipmaps[i]);
// }
// 
///////////////////////////////////////////////////////////////////////////////
// 
// Loading an uncompressed texture:
//
// CDDSImage image;
// GLuint texobj;
//
// image.load("uncompressed.dds");
//
// glGenTextures(1, &texobj);
// glEnable(GL_TEXTURE_2D);
// glBindTexture(GL_TEXTURE_2D, texobj);
//
// glTexImage2D(GL_TEXTURE_2D, 0, image.components, image[0].width, 
//     image[0].height, 0, image.format, GL_UNSIGNED_BYTE, image[0]);
//
// for (int i = 0; i < image[0].mipmaps.size(); i++)
// {
//     glTexImage2D(GL_TEXTURE_2D, i+1, image.components, 
//         image[0].mipmaps[i].width, image[0].mipmaps[i].height, 
//         0, image.format, GL_UNSIGNED_BYTE, image[0].mipmaps[i]);
// }
//
///////////////////////////////////////////////////////////////////////////////
// 
// Loading an uncompressed cubemap texture:
//
// CDDSImage image;
// GLuint texobj;
// GLenum target;
//
// image.load("cubemap.dds");
//
// glGenTextures(1, &texobj);
// glEnable(GL_TEXTURE_CUBE_MAP_ARB);
// glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, texobj);
//
// for (int n = 0; n < 6; n++)
// {
//     target = GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB+n;
//
//     glTexImage2D(target, 0, image.components, image[n].width, 
//         image[n].height, 0, image.format, GL_UNSIGNED_BYTE, image[n]);
//
//     for (int i = 0; i < image[n].mipmaps.size(); i++)
//     {
//         glTexImage2D(target, i+1, image.components, 
//             image[n].mipmaps[i].width, image[n].mipmaps[i].height, 
//             0, image.format, GL_UNSIGNED_BYTE, image[n].mipmaps[i]);
//     }
// }
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <assert.h>
#include "nv_dds.h"

#ifdef __APPLE__
inline void ENDIAN_WORD(void *pWord)
{
    unsigned char *pBytes = (unsigned char *)pWord;
    unsigned char temp;
    
    temp = pBytes[0];
    pBytes[0] = pBytes[1];
    pBytes[1] = temp;
}



inline void ENDIAN_DWORD(void *pWord)
{
    unsigned char *pBytes = (unsigned char *)pWord;
    unsigned char temp;
    
    // Swap outer bytes
    temp = pBytes[3];
    pBytes[3] = pBytes[0];
    pBytes[0] = temp;
    
    // Swap inner bytes
    temp = pBytes[1];
    pBytes[1] = pBytes[2];
    pBytes[2] = temp;
}
#else
#define ENDIAN_WORD 
#define ENDIAN_DWORD 
#endif    


// CSurface implementation

CSurface::CSurface()
{
    pixels = NULL;
    width = 0;
    height = 0;
    size = 0;
}

CSurface::CSurface(int w, int h, int imgsize)
{
    pixels = NULL;
    create(w, h, imgsize);
}

CSurface::CSurface(const CSurface &copy)
{
    pixels = NULL;
    width = 0;
    height = 0;
    size = 0;

    if (copy.pixels)
    {
        size = copy.size;
        width = copy.width;
        height = copy.height;
        pixels = (char*)malloc(size);
        memcpy(pixels, copy.pixels, copy.size);
    }
}

CSurface &CSurface::operator= (const CSurface &rhs)
{
    clear();
    if (this != &rhs)
    {
        if (rhs.pixels)
        {
            size = rhs.size;
            width = rhs.width;
            height = rhs.height;

            pixels = (char*)malloc(size);
            memcpy(pixels, rhs.pixels, size);
        }
    }

    return *this;
}

CSurface::operator char*(void)
{ 
    return pixels; 
}

void CSurface::create(int w, int h, int imgsize)
{
    clear();
    width = w;
    height = h;
    size = imgsize;
    pixels = (char*)malloc(imgsize);
}

CSurface::~CSurface()
{
    clear();
}

void CSurface::clear()
{
    if (pixels)
    {
        //delete [] pixels;
        free(pixels);
        pixels = 0;
    }
}

// CImage implementation

CImage::CImage()
{
    width = 0;
    height = 0;
    size = 0;
}

CImage::~CImage()
{
    mipmaps.clear();
}

CImage::CImage(int w, int h, int imgSize)
{
    width = w;
    height = h;
    size = imgSize;
    image.create(w, h, imgSize);
}

CImage::CImage(const CImage &copy)
{
    image.clear();
    width = copy.width;
    height = copy.height;
    size = copy.size;
    mipmaps.clear();

    image = copy.image;

    for (unsigned int i = 0; i < copy.mipmaps.size(); i++)
        mipmaps.push_back(copy.mipmaps[i]);
}

CImage::operator char*(void)
{ 
    return image; 
}

CDDSImage::CDDSImage()
{
    format = 0;
    components = 0;
    compressed = false;
    cubemap = false;
}

CDDSImage::~CDDSImage()
{
}

int CDDSImage::size_dxtc(int width, int height)
{
    return ((width+3)/4)*((height+3)/4)*
        (format == GL_COMPRESSED_RGBA_S3TC_DXT1_EXT ? 8 : 16);   
}

int CDDSImage::size_rgb(int width, int height)
{
    return width*height*components;   
}

bool CDDSImage::load(string filename)
{
	DDS_HEADER ddsh;
	char filecode[4];
	FILE *fp;
    int width, height;
    int (CDDSImage::*sizefunc)(int, int);

    clear();
    
	fp = fopen(filename.data(), "rb");
	if (fp == NULL)
		return false;

	fread(filecode, 1, 4, fp);
	if (strncmp(filecode, "DDS ", 4) != 0)
	{
		fclose(fp);
		return false;
	}

	fread(&ddsh, sizeof(ddsh), 1, fp);
    
    /*
     struct DDS_HEADER
     {
         unsigned long dwSize;
         unsigned long dwHeaderFlags;
         unsigned long dwHeight;
         unsigned long dwWidth;
         unsigned long dwPitchOrLinearSize;
         unsigned long dwDepth;
         unsigned long dwMipMapCount;
         unsigned long dwReserved1[11];
         DDS_PIXELFORMAT ddspf;
         unsigned long dwSurfaceFlags;
         unsigned long dwCubemapFlags;
         unsigned long dwReserved2[3];
     };
     struct DDS_PIXELFORMAT
     {
         unsigned long dwSize;
         unsigned long dwFlags;
         unsigned long dwFourCC;
         unsigned long dwRGBBitCount;
         unsigned long dwRBitMask;
         unsigned long dwGBitMask;
         unsigned long dwBBitMask;
         unsigned long dwABitMask;
     };
     
     
*/     
    // ENDIAN_DWORD

#ifdef __APPLE__
    ENDIAN_DWORD(&ddsh.dwSize);
    ENDIAN_DWORD(&ddsh.dwHeaderFlags);
    ENDIAN_DWORD(&ddsh.dwHeight);
    ENDIAN_DWORD(&ddsh.dwWidth);
    ENDIAN_DWORD(&ddsh.dwPitchOrLinearSize);
    ENDIAN_DWORD(&ddsh.dwDepth);
    ENDIAN_DWORD(&ddsh.dwMipMapCount);

    ENDIAN_DWORD(&ddsh.dwReserved1[0]);
    ENDIAN_DWORD(&ddsh.dwReserved1[1]);
    ENDIAN_DWORD(&ddsh.dwReserved1[2]);
    ENDIAN_DWORD(&ddsh.dwReserved1[3]);
    ENDIAN_DWORD(&ddsh.dwReserved1[4]);
    ENDIAN_DWORD(&ddsh.dwReserved1[5]);                             
    ENDIAN_DWORD(&ddsh.dwReserved1[6]);    
    ENDIAN_DWORD(&ddsh.dwReserved1[7]);
    ENDIAN_DWORD(&ddsh.dwReserved1[8]);
    ENDIAN_DWORD(&ddsh.dwReserved1[9]);
    ENDIAN_DWORD(&ddsh.dwReserved1[10]);

    ENDIAN_DWORD(&ddsh.ddspf.dwSize);
    ENDIAN_DWORD(&ddsh.ddspf.dwFlags);
    ENDIAN_DWORD(&ddsh.ddspf.dwFourCC);
    ENDIAN_DWORD(&ddsh.ddspf.dwRGBBitCount);
    ENDIAN_DWORD(&ddsh.ddspf.dwRBitMask);
    ENDIAN_DWORD(&ddsh.ddspf.dwGBitMask);
    ENDIAN_DWORD(&ddsh.ddspf.dwBBitMask);
    ENDIAN_DWORD(&ddsh.ddspf.dwABitMask);
    
    ENDIAN_DWORD(&ddsh.dwSurfaceFlags);
    ENDIAN_DWORD(&ddsh.dwCubemapFlags);
    ENDIAN_DWORD(&ddsh.dwReserved2[0]);
    ENDIAN_DWORD(&ddsh.dwReserved2[1]);
    ENDIAN_DWORD(&ddsh.dwReserved2[2]);
#endif
    
    if (ddsh.dwCubemapFlags & DDS_CUBEMAP_ALLFACES)
        cubemap = true;

    if (ddsh.ddspf.dwFlags & DDS_FOURCC) 
    {
	    switch(ddsh.ddspf.dwFourCC)
	    {
            case FOURCC_DXT1:
                format = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
                components = 3;
                compressed = true;
                break;
            case FOURCC_DXT3:
                format = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
                components = 4;
                compressed = true;
                break;
            case FOURCC_DXT5:
                format = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
                components = 4;
                compressed = true;
                break;
            default:
                fclose(fp);
                return false;
	    }
    }
    else if (ddsh.ddspf.dwFlags == DDS_RGBA && ddsh.ddspf.dwRGBBitCount == 32)
    {
        format = GL_BGRA_EXT; 
        compressed = false;
        components = 4;
    }
    else if (ddsh.ddspf.dwFlags == DDS_RGB  && ddsh.ddspf.dwRGBBitCount == 32)
    {
        format = GL_BGRA_EXT; 
        compressed = false;
        components = 4;
    }
    else if (ddsh.ddspf.dwFlags == DDS_RGB  && ddsh.ddspf.dwRGBBitCount == 24)
    {
        format = GL_BGR_EXT; 
        compressed = false;
        components = 3;
    }
    else 
    {
        fclose(fp);
        return false;
    }

    width = ddsh.dwWidth;
    height = ddsh.dwHeight;
    
    sizefunc = (compressed ? &CDDSImage::size_dxtc : &CDDSImage::size_rgb);

    for (int n = 0; n < (cubemap ? 6 : 1); n++)
    {
        
        CImage img(width, height, (this->*sizefunc)(width, height));
        fread(img, 1, img.size, fp);
        flip(img, img.width, img.height, img.size);

        int h = height >> 1;
        int w = width >> 1;
    
        for (unsigned int i = 0; i < ddsh.dwMipMapCount && (w || h); i++)
        {
            CSurface mipmap(h, w, (this->*sizefunc)(w, h));
            fread(mipmap, 1, mipmap.size, fp);
            flip(mipmap, mipmap.width, mipmap.height, mipmap.size);

            img.mipmaps.push_back(mipmap);

            h >>= 1;
            w >>= 1;
        }
        images.push_back(img);
    }

    if (cubemap)
    {
        CImage tmp;
        tmp = images[3];
        images[3] = images[2];
        images[2] = tmp;
    }
    
    fclose(fp);

	return true;
}

void swap(void *byte1, void *byte2, int size)
{
    unsigned char *tmp = (unsigned char*)malloc(size);

    memcpy(tmp, byte1, size);
    memcpy(byte1, byte2, size);
    memcpy(byte2, tmp, size);

    delete [] tmp;
}

void flip_blocks_dxtc1(DXTColBlock *line, int numBlocks)
{
    DXTColBlock *curblock = line;

    for (int i = 0; i < numBlocks; i++)
    {
        swap(&curblock->row[0], &curblock->row[3], sizeof(unsigned char));
        swap(&curblock->row[1], &curblock->row[2], sizeof(unsigned char));

        curblock++;
    }
}

void flip_blocks_dxtc35(DXTColBlock *line, int numBlocks)
{
    DXTColBlock *curblock = line;
    DXTAlphaBlock *alphablock;

    for (int i = 0; i < numBlocks; i++)
    {
        alphablock = (DXTAlphaBlock*)curblock;

        swap(&alphablock->row[0], &alphablock->row[3], sizeof(unsigned short));
        swap(&alphablock->row[1], &alphablock->row[2], sizeof(unsigned short));

        curblock++;

        swap(&curblock->row[0], &curblock->row[3], sizeof(unsigned char));
        swap(&curblock->row[1], &curblock->row[2], sizeof(unsigned char));

        curblock++;
    }
}

void CDDSImage::flip(char *image, int width, int height, int size)
{
    int linesize;

    if (!compressed)
    {
        linesize = size / height;

        char *top = image;
        char *bottom = image + (size-linesize);
    
        for (int i = 0; i < (height >> 1)-1; i++)
        {
            swap(bottom, top, linesize);

            top += linesize;
            bottom -= linesize;
        }
    
    }
    else
    {
        void (*flipblocks)(DXTColBlock*, int);
    	int xblocks = width / 4;
    	int yblocks = height / 4;
        int blocksize;

        switch (format)
        {
            case GL_COMPRESSED_RGBA_S3TC_DXT1_EXT: 
                blocksize = 8;
                flipblocks = flip_blocks_dxtc1; 
                break;
            case GL_COMPRESSED_RGBA_S3TC_DXT3_EXT: 
                blocksize = 16;
                flipblocks = flip_blocks_dxtc35; 
                break;
            case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT: 
                blocksize = 16;
                flipblocks = flip_blocks_dxtc35; 
                break;
            default:
                return;
        }

        linesize = xblocks * blocksize;

    	DXTColBlock *top;
    	DXTColBlock *bottom;
    
	    for (int j = 0; j < (yblocks >> 1); j++)
	    {
    		top = (DXTColBlock*)((unsigned long)image + j * linesize );
    		bottom = (DXTColBlock*)((unsigned long)image + (((yblocks-j)-1) * linesize));

            flipblocks(top, xblocks);
            flipblocks(bottom, xblocks);

            swap(bottom, top, linesize);
        }
    }

}    

void CDDSImage::clear()
{
    components = 0;
    format = 0;
    compressed = false;
    cubemap = false;

    images.clear();
}

CImage &CDDSImage::operator[](int i)
{ 
    // make sure an image has been loaded
    assert(i < (int)images.size());

    return images[i]; 
}
